package abc.da.weaving.aspectinfo;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import polyglot.types.SemanticException;
import abc.da.ast.DAAdviceDecl;
import abc.weaving.aspectinfo.AbstractAdviceDecl;
import abc.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.GlobalAspectInfo;

/**
 * An enhanced version of {@link GlobalAspectInfo} which keeps track of advice names
 * and dependencies.
 * @author Eric Bodden
 */
public class DAGlobalAspectInfo extends GlobalAspectInfo {

	protected Map<String,String> adviceMethodNameToAdviceShortName = new HashMap<String,String>();
	protected Set<AdviceDependency> adviceDependencies = new HashSet<AdviceDependency>();
	
	/**
	 * Registers a new advice dependency.
	 */
	public void addAdviceDependency(AdviceDependency ad) {
		adviceDependencies.add(ad);
	}
	
	/**
	 * Returns the unmodifiable set of all advice dependencies.
	 */
	public Set<AdviceDependency> getAdviceDependencies() {
		return Collections.unmodifiableSet(adviceDependencies);		
	}
	
	/**
	 * Registers a human-readable name for an internal advice name.
	 * Both names must not be qualified.
	 */
	public void registerHumanReadableNameForAdviceName(Aspect owningAspect, String internalAdviceName, String humanReadableName) {
		if(internalAdviceName.contains(".")) {
			throw new IllegalArgumentException("Internal advice name may not be qualified!");
		}
		//qualify
		internalAdviceName = owningAspect.getName() + "." + internalAdviceName;
		if(adviceMethodNameToAdviceShortName.containsKey(internalAdviceName)) {
			throw new RuntimeException("already registered!");
		}
		
		
		if(humanReadableName.contains(".")) {
			throw new IllegalArgumentException("Human readable advice name may not be qualified!");
		}
		//qualify
		humanReadableName = owningAspect.getName() + "." + humanReadableName;
		adviceMethodNameToAdviceShortName.put(internalAdviceName, humanReadableName);
	}
	
	/**
	 * Returns for a fully qualified internal advice name the fully qualified human-readable name
	 * if such a name was previously registered. Otherwise, the original name is returned. 
	 */
	public String replaceForHumanReadableName(String internalAdviceName) {
		String hrn = adviceMethodNameToAdviceShortName.get(internalAdviceName);
		if(hrn!=null) {
			return hrn;
		} else {
			return internalAdviceName;
		}
	}
	
	/**
	 * Performs a consistency check on dependent advice declarations. Each dependent advice must be mentioned in at
	 * least one dependency declaration. Also each dependency declaration may only refer to existing dependent advice. 
	 */
	public void consistencyCheckForDependentAdvice() {
		Set<String> qualifiedDependentAdviceNamesDeclared = new HashSet<String>();
		for (AbstractAdviceDecl ad : getAdviceDecls()) {
			if(ad.getFlags().intersects(DAAdviceDecl.DEPENDENT)) {
				qualifiedDependentAdviceNamesDeclared.add(replaceForHumanReadableName(ad.getQualifiedAdviceName()));
			}
		}
		
		Set<String> qualifiedDependentAdviceNamesFound = new HashSet<String>();		
		//check that each advice mentioned really exists in code
		for (AdviceDependency dep : getAdviceDependencies()) {
			for (String qualifiedAdviceName  : dep.adviceNames()) {
				if(!qualifiedDependentAdviceNamesDeclared.contains(qualifiedAdviceName)) {
					throw new RuntimeException(new SemanticException("Advice with name '"+qualifiedAdviceName+
							"' mentioned in dependency is not found in aspect "+dep.getContainer()+"!", dep.getPosition()));
				}
				qualifiedDependentAdviceNamesFound.add(qualifiedAdviceName);
			}
			
			if(dep.strongAdviceNameToVars.isEmpty()) {
				throw new RuntimeException(new SemanticException("Dependency group declares no strong advice!", dep.getPosition()));
			}
		}
		
		for (AbstractAdviceDecl ad : getAdviceDecls()) {
			if(ad.getFlags().intersects(DAAdviceDecl.DEPENDENT)) {
				String qualified = replaceForHumanReadableName(ad.getQualifiedAdviceName());
				if(!qualifiedDependentAdviceNamesFound.contains(qualified)) {
					throw new RuntimeException(new SemanticException("Dependent advice '"+qualified+"' is never " +
							"referenced in any dependency declaration.", ad.getPosition()));
				}
			}
		}
	}
	
}
