package abc.da.weaving.weaver.depadviceopt.ds;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import polyglot.util.ErrorInfo;
import polyglot.util.Position;
import polyglot.util.StdErrorQueue;
import soot.Local;
import soot.PointsToSet;
import soot.RefLikeType;
import soot.Scene;
import soot.SootMethod;
import soot.Unit;
import soot.jimple.InvokeExpr;
import soot.jimple.Stmt;
import soot.jimple.toolkits.pointer.FullObjectSet;
import soot.tagkit.Host;
import soot.tagkit.LineNumberTag;
import soot.tagkit.SourceLnNamePosTag;
import soot.tagkit.SourceLnPosTag;
import abc.da.ast.DAAdviceDecl;
import abc.da.weaving.aspectinfo.DAGlobalAspectInfo;
import abc.main.Main;
import abc.tm.weaving.weaver.tmanalysis.query.WeavableMethods;
import abc.weaving.aspectinfo.AbstractAdviceDecl;
import abc.weaving.aspectinfo.AdviceDecl;
import abc.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.Formal;
import abc.weaving.matching.AdviceApplication;
import abc.weaving.matching.MethodAdviceList;
import abc.weaving.residues.NeverMatch;
import abc.weaving.residues.Residue;
import abc.weaving.residues.ResidueBox;
import abc.weaving.tagkit.InstructionShadowTag;

/**
 * Represents a joinpoint shadow in Jimple. Useful for shadow-based optimizations.
 * @author Eric Bodden
 */
public class Shadow {
	
	protected static DAGlobalAspectInfo gai;
	
	/**
	 * The unique numeric ID of this shadow.
	 */
	protected final int shadowId;
	
	/**
	 * The advice declaration that caused this shadow to be woven. 
	 */
	protected final AdviceDecl adviceDecl;
	
	/**
	 * The method containing this shadow, 
	 */
	protected final SootMethod container;
		
	/**
	 * The position at which this shadow occurs in code.
	 */
	protected Position pos;
	
	/**
	 * The {@link ResidueBox} for this shadow. Can be used to modify the shadow's residue.
	 */
	protected final ResidueBox outerResidueBox;
	
	/**
	 * The list of variables bound by this shadows, in the order of declaration. 
	 */
	protected final List<String> variableOrder;
	
	/**
	 * A mapping from advice formal names to {@link Local}s that bind these formals in code. 
	 */
	protected final Map<String, Local> adviceFormalNameToSootLocal;

	/**
	 * A mapping from advice formal names to points-to sets of the corresponding {@link Local}s.
	 */
	protected final Map<String, PointsToSet> adviceFormalNameToPointsToSet;
	
	/**
	 * The set of formals of this shadow that bind non-{@link RefLikeType}s.
	 */
	protected final Set<String> primitiveFormalNames;
	
	private Shadow(int shadowId, AdviceDecl adviceDecl, SootMethod container, Position pos, Map<String, Local> adviceFormalNameToSootLocal, ResidueBox outerResidueBox) {
		this.shadowId = shadowId;
		this.adviceDecl = adviceDecl;
		this.container = container;
		this.adviceFormalNameToSootLocal = adviceFormalNameToSootLocal;
		this.pos = pos;
		this.outerResidueBox = outerResidueBox;
		this.adviceFormalNameToPointsToSet = new HashMap<String, PointsToSet>();
		List<Formal> formals = adviceDecl.getFormals();
		this.variableOrder = new ArrayList<String>(formals.size());
		this.primitiveFormalNames = new HashSet<String>();
		for (Formal formal : formals) {
			this.variableOrder.add(formal.getName());
			if(!(formal.getType().getSootType() instanceof RefLikeType)) {
				primitiveFormalNames.add(formal.getName());
			}
		}
	}
	
	/**
	 * Returns the aspect that declares the advice that caused this shadow to be woven.
	 */
	public Aspect declaringAspect() {
		return adviceDecl.getAspect();
	}
	
	/**
	 * Disables this shadow by setting its {@link Residue} to {@link NeverMatch}.
	 */
	public void disable() {
		outerResidueBox.setResidue(NeverMatch.v());
	}
	
	/**
	 * Returns <code>true</code> if this shadow's {@link Residue} is anything else but
	 * {@link NeverMatch}.
	 */
	public boolean isEnabled() {
		return !NeverMatch.neverMatches(outerResidueBox.getResidue());
	}
	
	/**
	 * Returns the names of variables bound by this shadow in the order of declaration. 
	 */
	public List<String> variableNames() {
		return variableOrder;
	}
	
	/**
	 * Returns <code>true</code> if the advice that caused this shadow to be woven is
	 * a dependent advice.
	 */
	public boolean belongsToDependentAdvice() {
		return adviceDecl.getFlags().intersects(DAAdviceDecl.DEPENDENT);
	}
	
	/**
	 * Returns the (potentially human-readable) name of the dependent advice that
	 * caused this shadow to be woven. May only be called if {@link #belongsToDependentAdvice()} 
	 * returns <code>true</code>.
	 */
	public String dependentAdviceName() {
		if(!belongsToDependentAdvice()) {
			throw new IllegalStateException("Shadow does not belong to a dependent advice.");
		} else {
			return gai.replaceForHumanReadableName(adviceDecl.getQualifiedAdviceName());
		}
	}
	
	/**
	 * Computes and returns the {@link PointsToSet} for variable
	 * var bound by this shadow. If this shadow does not bind var,
	 * or if it binds var to a primitive value, {@link FullObjectSet} is returned.
	 */
	public PointsToSet pointsToSetOf(String var) {
		if(!variableNames().contains(var) || primitiveFormalNames.contains(var)) {
			return FullObjectSet.v();
		}
		PointsToSet pts = adviceFormalNameToPointsToSet.get(var);
		if(pts==null) {
			Local l = adviceFormalNameToSootLocal.get(var);
			assert l!=null;
			pts = Scene.v().getPointsToAnalysis().reachingObjects(l);
			adviceFormalNameToPointsToSet.put(var, pts);
		}
		return pts;
	}
	
	@Override
	public String toString() {
		String ret = "";
		
		if(pos!=null) {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			PrintStream ps = new PrintStream(bos);		
			StdErrorQueue tmpQueue = new StdErrorQueue(ps,80,"tmp");
			tmpQueue.enqueue(ErrorInfo.WARNING, "", pos);
			tmpQueue.flush();
			String markers = bos.toString();
			markers = markers.replace("Warning --", "");
			markers = markers.replace("\n\n", "");
			ret += markers + "\n\n";
		}

		ret += "shadowId:     " + shadowId + "\n";
		ret += "advice:       " + adviceDecl  + "\n";
		if(belongsToDependentAdvice()) {
			ret += "dep. advice:  " + dependentAdviceName() + "\n";
		}
		ret += "in method:    " + container  + "\n";
		if(pos!=null) {
		    ret += "position:     " + pos + "\n";
		}
		ret += "variables:    " + adviceFormalNameToSootLocal + "\n";
		ret += "residue:      " + outerResidueBox.getResidue() + "\n";
		
		return ret;
	}
		
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((adviceDecl == null) ? 0 : adviceDecl.hashCode());
		result = prime * result + shadowId;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Shadow other = (Shadow) obj;
		if (adviceDecl == null) {
			if (other.adviceDecl != null)
				return false;
		} else if (!adviceDecl.equals(other.adviceDecl))
			return false;
		if (shadowId != other.shadowId)
			return false;
		return true;
	}

	/**
	 * Creates and returns all active shadows in all weavable methods.
	 * @param shadowsOfDependentAdviceOnly If <code>true</code>, only shadows of dependent
	 * advice are returned.
	 */
	public static Set<Shadow> allActiveShadows(boolean shadowsOfDependentAdviceOnly) {
		return findActiveShadowsInMethod(WeavableMethods.v().getAll(), shadowsOfDependentAdviceOnly);
	}

	/**
	 * Creates and returns all active shadows in all weavable methods reachable from the
	 * program's entry points.
	 * @param shadowsOfDependentAdviceOnly If <code>true</code>, only shadows of dependent
	 * advice are returned.
	 */
	public static Set<Shadow> reachableActiveShadows(boolean shadowsOfDependentAdviceOnly) {
		if(!Scene.v().hasCallGraph()) {
			throw new IllegalStateException("No callgraph present.");
		}
		return findActiveShadowsInMethod(WeavableMethods.v().getReachable(Scene.v().getCallGraph()), shadowsOfDependentAdviceOnly);
	}

	protected static Set<Shadow> findActiveShadowsInMethod(Set<SootMethod> methods, boolean shadowsOfDependentAdviceOnly) {
		gai = (DAGlobalAspectInfo) Main.v().getAbcExtension().getGlobalAspectInfo();
		
		Map<SootMethod,AdviceDecl> adviceMethodToAdviceDecl = new HashMap<SootMethod, AdviceDecl>();
		for (AbstractAdviceDecl aad : (List<AbstractAdviceDecl>)gai.getAdviceDecls()) {
			//only care about "real" advice declarations
			if(aad instanceof AdviceDecl) {
				AdviceDecl ad = (AdviceDecl) aad;
				SootMethod adviceMethod = ad.getImpl().getSootMethod();
				adviceMethodToAdviceDecl.put(adviceMethod, ad);
			}			
		}

		Set<Shadow> shadows = new HashSet<Shadow>();
		for (SootMethod m : methods) {

			Map<Integer,Map<String,AdviceApplication>> shadowIdToAdviceNameToAdviceApplication = new HashMap<Integer,Map<String,AdviceApplication>>();
	        MethodAdviceList adviceList = gai.getAdviceList(m);
	        //if there are any advice applications within that method
	        if(adviceList!=null) {
	        	//build a mapping from shadow ID and advice name to the corresponding advice application
		        List<AdviceApplication> applications = adviceList.allAdvice();
		        for (AdviceApplication aa : applications) {	    
		        	Map<String, AdviceApplication> adviceNameToAA = shadowIdToAdviceNameToAdviceApplication.get(aa.shadowmatch.shadowId);
		        	if(adviceNameToAA==null) {
		        		adviceNameToAA = new HashMap<String, AdviceApplication>();
		        		shadowIdToAdviceNameToAdviceApplication.put(aa.shadowmatch.shadowId, adviceNameToAA);
		        	}		        	
		        	adviceNameToAA.put(aa.advice.getQualifiedAdviceName(), aa);
		        }
		
		        for (Unit u : m.getActiveBody().getUnits()) {
					Stmt s = (Stmt) u;
					if(s.containsInvokeExpr()) {
						InvokeExpr ie = s.getInvokeExpr();
						AdviceDecl ad = adviceMethodToAdviceDecl.get(ie.getMethod());
						if (ad!=null) {
							
							//the called method is an advice method; we just found a shadow!						
							if(!shadowsOfDependentAdviceOnly || ad.getFlags().intersects(DAAdviceDecl.DEPENDENT)) {

								InstructionShadowTag shadowTag = (InstructionShadowTag) s.getTag(InstructionShadowTag.NAME);
								int shadowId = shadowTag.value();
								String qualifiedAdviceName = s.getInvokeExpr().getMethod().getDeclaringClass().getName()
									+ "." + s.getInvokeExpr().getMethod().getName();
								AdviceApplication aa = shadowIdToAdviceNameToAdviceApplication.get(shadowId).get(qualifiedAdviceName);
								if(!NeverMatch.neverMatches(aa.getResidue())) {
									//advice is still active
									
									Map<String,Local> adviceFormalToSootLocal = new HashMap<String, Local>();
									int argIndex = 0;
									for (Formal formal : ad.getFormals()) {
										adviceFormalToSootLocal.put(formal.getName(), (Local)ie.getArg(argIndex));
										argIndex++;
									}
									
									Position pos = extractPosition(aa.shadowmatch.getHost());
									ResidueBox rbox = (ResidueBox) aa.getResidueBoxes().get(0); 
									Shadow shadow = new Shadow(
											shadowId,
											ad,
											m,
											pos,
											adviceFormalToSootLocal,
											rbox
									);
									shadows.add(shadow);
									
								}
							}
							
						}
					}
		        }
			}
		}
		
		return shadows;
	}

	/**
	 * Returns the position at which the shadow was woven.
	 */
	public Position getPosition() {
		return pos;
	}
	
	private static Position extractPosition(Host host) {
    	if(host.hasTag("SourceLnPosTag")) {
    		SourceLnPosTag tag = (SourceLnPosTag) host.getTag("SourceLnPosTag");
    		String fileName = "";
    		if(tag instanceof SourceLnNamePosTag) {
				SourceLnNamePosTag nameTag = (SourceLnNamePosTag) tag;
    			fileName = nameTag.getFileName();
    		}    		
    		return new Position(fileName,tag.startLn(),tag.startPos(),tag.endLn(),tag.endPos());
    	} else if(host.hasTag("LineNumberTag")) {
    		LineNumberTag tag = (LineNumberTag) host.getTag("LineNumberTag");
    		return new Position("",tag.getLineNumber());
    	} else {
    		return null;
    	}
    }

	/**
	 * Returns the advice declaration that caused this shadow to be woven.
	 */
	public AdviceDecl getAdviceDecl() {
		return adviceDecl;
	}    

}
