package abc.tm.weaving.weaver.tmanalysis;

import java.util.HashSet;
import java.util.Set;

import polyglot.util.ErrorInfo;
import abc.da.ast.DAAdviceDecl;
import abc.da.weaving.aspectinfo.AdviceDependency;
import abc.da.weaving.weaver.depadviceopt.DependentAdviceQuickCheck;
import abc.main.Main;
import abc.tm.weaving.aspectinfo.TMAdviceDependency;
import abc.tm.weaving.aspectinfo.TMGlobalAspectInfo;
import abc.tm.weaving.aspectinfo.TraceMatch;
import abc.weaving.aspectinfo.AbstractAdviceDecl;
import abc.weaving.matching.AdviceApplication;

/**
 * Customized version of {@link DependentAdviceQuickCheck} that reports warnings in terms
 * of tracematches.
 * @author Eric Bodden
 */
public class TMQuickCheck extends DependentAdviceQuickCheck {

	protected Set<AbstractAdviceDecl> dependentTmAdvice;
	
	protected Set<TMAdviceDependency> nonFulfilledDependencies;
	
	/** 
	 * Populates {@link #dependentTmAdvice} with the set of all dependent advice declarations
	 * that belong to a tracematch. Also, removes all transitions from all tracematches
	 * where the transition's symbol never matches. Then invokes the actual analysis.
	 */
	@Override
	public boolean analyze() {

		final TMGlobalAspectInfo gai = (TMGlobalAspectInfo) Main.v().getAbcExtension().getGlobalAspectInfo();
		final Set<AdviceDependency> adviceDependencies = gai.getAdviceDependencies();
		final Set<AdviceDependency> tmAdviceDependencies = new HashSet<AdviceDependency>(); 
		for (AdviceDependency adviceDependency : adviceDependencies) {
			if(adviceDependency instanceof TMAdviceDependency) {
				tmAdviceDependencies.add(adviceDependency);
			}
		}
		
		dependentTmAdvice = new HashSet<AbstractAdviceDecl>();
		
		for (AbstractAdviceDecl aad : gai.getAdviceDecls()) {
			boolean isDependent = aad.getFlags().intersects(DAAdviceDecl.DEPENDENT);
			if(isDependent) {
				for (AdviceDependency ad : tmAdviceDependencies) {
					if(ad.containsAdviceNamed(aad.getQualifiedAdviceName())) {
						dependentTmAdvice.add(aad);
					}
				}
			}
		}
		
		//remove all transition from the state machine that never match
		for (TraceMatch tm: gai.getTraceMatches()) {
			tm.removeNonMatchingTransitions();			
		}
		
		nonFulfilledDependencies = new HashSet<TMAdviceDependency>();
		
		
		/*
		 * ACTUALLY CONDUCT THE ANALYSIS
		 */
		boolean result = super.analyze();

		/*
		 * We report a warning if the quick check removes a tracematch entirely, i.e. if
		 * all its dependencies are not fulfilled.
		 */
		Set<AdviceDependency> fulfilled = new HashSet<AdviceDependency>(adviceDependencies);
		fulfilled.removeAll(nonFulfilledDependencies);
		//for each non-fulfilled dependency
		for (AdviceDependency ad : nonFulfilledDependencies) {
			if(ad instanceof TMAdviceDependency) {
				TMAdviceDependency tmad = (TMAdviceDependency) ad;
				TraceMatch tm = tmad.getTracematch();
				boolean found = false;
				//see if another dependency of the same tm is fulfilled
				for (AdviceDependency fulfilledAd : fulfilled) {
					if(ad instanceof TMAdviceDependency) {
						TMAdviceDependency tmad2 = (TMAdviceDependency) fulfilledAd;
						TraceMatch tm2 = tmad2.getTracematch();
						if(tm==tm2) {
							found = true;
							break;
						}
					}
				}
				//if not, warn
				if(!found) {
					Main.v().getAbcExtension().forceReportError(
							ErrorInfo.WARNING,
							"This tracematch can never match, as determined by the Quick Check. All its shadows will be disabled.",
							tm.getPosition());
				}
			}
		}		
		
		return result;
	}
	

	
	/** 
	 * Does not warn for each single advice of a tracematch.
	 */
	@Override
	protected void warn(AdviceApplication aa) {
		//don't warn about each single advice for a tracematch
		if(!dependentTmAdvice.contains(aa.advice))
			super.warn(aa);
		
	}
	
	/** 
	 * Warns for a dependency induced by a tracematch.
	 */
	@Override
	protected void warn(AdviceDependency nonFulfilled) {
		if(nonFulfilled instanceof TMAdviceDependency) {
			this.nonFulfilledDependencies.add((TMAdviceDependency) nonFulfilled);
		} else {
			super.warn(nonFulfilled);
		}
	}

}
