/* abc - The AspectBench Compiler
 * Copyright (C) 2007 Eric Bodden
 *
 * This compiler is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This compiler is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this compiler, in the file LESSER-GPL;
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package abc.tm.weaving.weaver.tmanalysis.stages;

import java.util.Iterator;

import soot.Body;
import soot.MethodOrMethodContext;
import soot.PackManager;
import soot.Scene;
import soot.SootMethod;
import soot.Unit;
import soot.jimple.toolkits.callgraph.CallGraph;
import abc.tm.weaving.weaver.tmanalysis.callgraph.AbstractedCallGraph;
import abc.tm.weaving.weaver.tmanalysis.callgraph.NodePredicate;
import abc.tm.weaving.weaver.tmanalysis.stages.TMShadowTagger.SymbolShadowTag;
import abc.tm.weaving.weaver.tmanalysis.util.ISymbolShadow;
import abc.weaving.aspectinfo.MethodCategory;

/**
 * This stage does not actually perform any real analysis. It merely applies the <i>cg</i> phase, constructing a call graph
 * and points-to sets and afterwards abstracts the call graph, pruning subgraphs which contain no methods of interest. 
 * However, afterwards it disables all unrechable shadows, which might lead to the fact that there are actually no enabled
 * shadows remaining (in the case where all active shadows were unreachable).
 *
 * @author Eric Bodden
 */
public class CallGraphAbstraction {

    /** 
     * predicate used for call graph  abstraction
     * @see TaggedMethods
     */
    protected final NodePredicate ONLY_METHODS_WITH_MATCHED_UNITS = new TaggedMethods();
    
	/**
	 * {@inheritDoc}
	 */
	public CallGraph abstractedCallGraph() {
        //set a main class from the options if none is set yet
        Scene.v().setMainClassFromOptions();
        
    	//build call graph
    	PackManager.v().getPack("cg").apply();

		CallGraph callGraph = Scene.v().getCallGraph();

        return new AbstractedCallGraph(callGraph, ONLY_METHODS_WITH_MATCHED_UNITS);
	}

    /**
     * This is a predicate over call graph nodes which accepts
     * only methods which contain at least one unit which is tagged
     * with matching tracematch symbols.
     */
	private class TaggedMethods implements NodePredicate {
		
		/** 
	     * @return <code>true</code> if the method is tagged with matching symbols or it 
	     * contains a unit that is tagged with matching symbols.
	     */
	    public boolean want(MethodOrMethodContext curr) {
	    	SootMethod method = curr.method();
	    	
	    	//explicitly has no effects on base code
	    	if(MethodCategory.noEffectsOnBaseCode(method)) {
	    		return false;
	    	}

	        if(method.hasActiveBody()) {
	            Body body = method.getActiveBody();
	            
	            for (Iterator iter = body.getUnits().iterator(); iter.hasNext();) {
	                Unit u = (Unit) iter.next();
	                //if we have a tag
	                if(u.hasTag(SymbolShadowTag.NAME)) {
	                	SymbolShadowTag tag = (SymbolShadowTag) u.getTag(SymbolShadowTag.NAME);
						//if any shadows in the tag are still enabled 
	                	for (ISymbolShadow match : tag.getAllMatches()) {
							if(match.isEnabled()) {
								return true;
							}
						}
	                }
	            }
	        }
	        
	        return false;
	    }

		/** 
		 * Returns <code>true</code> if the method can call back to weavable classes.
		 */
		public boolean visitChildren(MethodOrMethodContext curr) {
	    	SootMethod method = curr.method();
	    	
	    	//explicitly has no effects on base code
	    	return !MethodCategory.noEffectsOnBaseCode(method);
	    }
	}
	
	//singleton pattern
	
	protected static CallGraphAbstraction instance;

	private CallGraphAbstraction() {}
	
	public static CallGraphAbstraction v() {
		if(instance==null) {
			instance = new CallGraphAbstraction();
		}
		return instance;		
	}
	
	/**
	 * Frees the singleton object. 
	 */
	public static void reset() {
		instance = null;
	}


}
